# (C) 2023 GoodData Corporation
from enum import Enum
from pathlib import Path
from typing import Any, TypeVar

import attrs
from cattrs import structure


class GoodDataLdmType(Enum):
    PRIMARY_KEY = "primary_key"
    REFERENCE = "reference"
    DATE = "date"
    FACT = "fact"
    ATTRIBUTE = "attribute"
    LABEL = "label"


class GoodDataLabelType(Enum):
    TEXT = "TEXT"
    HYPERLINK = "HYPERLINK"
    GEO_LATITUDE = "GEO_LATITUDE"
    GEO_LONGITUDE = "GEO_LONGITUDE"


class GoodDataSortDirection(Enum):
    ASC = "ASC"
    DESC = "DESC"


DATE_GRANULARITIES = [
    "DAY",
    "WEEK",
    "MONTH",
    "QUARTER",
    "YEAR",
    "DAY_OF_WEEK",
    "DAY_OF_MONTH",
    "DAY_OF_YEAR",
    "WEEK_OF_YEAR",
    "MONTH_OF_YEAR",
    "QUARTER_OF_YEAR",
]
TIMESTAMP_GRANULARITIES = [
    "MINUTE",
    "HOUR",
    "MINUTE_OF_HOUR",
    "HOUR_OF_DAY",
]
T = TypeVar("T", bound="Base")

DBT_TARGET_DIR = Path("target")
DBT_PATH_TO_MANIFEST = DBT_TARGET_DIR / "manifest.json"


class DbtTests(Enum):
    PRIMARY_KEY = "dbt_constraints.primary_key"
    FOREIGN_KEY = "dbt_constraints.foreign_key"
    FOREIGN_KEY_REF = "pk_table_name"


DATETIME_DATA_TYPES = ["DATE", "TIMESTAMP", "TIMESTAMPTZ"]

TIMESTAMP_DATA_TYPES = ["TIMESTAMP", "TIMESTAMPTZ"]

NUMERIC_DATA_TYPES = [
    "NUMERIC",
]


@attrs.define
class Base:
    @classmethod
    def from_dict(cls: type[T], data: dict[str, Any]) -> T:
        """
        Creates object from dictionary.
        """
        return structure(data, cls)

    def to_dict(self) -> dict[str, Any]:
        """
        Converts object into dictionary.
        """
        return attrs.asdict(self)



# (C) 2023 GoodData Corporation
import argparse
from pathlib import Path

from gooddata_dbt.dbt.profiles import DbtProfiles

_CURR_DIR = Path(__file__).parent


def test_profiles():
    dbt_profiles = DbtProfiles(argparse.Namespace(profiles_dir=_CURR_DIR / "resources/dbt_profiles"))
    profiles = dbt_profiles.profiles
    assert len(profiles) == 1
    default_profile = profiles[0]
    assert default_profile.name == "default"
    assert len(default_profile.outputs) == 6



# (C) 2023 GoodData Corporation
import json
import logging
import os
import time
from pathlib import Path
from typing import Union

import attrs
import requests
import yaml
from cattrs import structure
from gooddata_sdk.utils import safeget
from requests import Response

from gooddata_dbt.dbt.environment import DbtCloudEnvironment


@attrs.define(auto_attribs=True, kw_only=True)
class DbtCloudBase:
    base_v2: str = attrs.field(init=False, default="https://cloud.getdbt.com/api/v2")
    base_v3: str = attrs.field(init=False, default="https://cloud.getdbt.com/api/v3")
    base_graphql: str = attrs.field(init=False, default="https://metadata.cloud.getdbt.com/graphql")

    graphql_applied_models = """
        query AppliedModels($environmentId: BigInt!, $first: Int!) {
            environment(id: $environmentId) {
              applied {
                models(first: $first) {
                  edges {
                    node {
                      name
                      uniqueId
                      materializedType
                      executionInfo {
                        lastSuccessRunId
                        executionTime
                        executeStartedAt
                      }
                    }
                  }
                }
              }
            }
          }
        """
    graphql_model_historical_runs = """
        query modelHistoricalRuns($environmentId: BigInt!, $modelId: String!, $first: Int!) {
          environment(id: $environmentId) {
            applied {
              modelHistoricalRuns(
                uniqueId: $modelId
                lastRunCount: $first
              ) {
                runId # Get historical results for a particular model
                runGeneratedAt
                executionTime # View build time across runs
                status
              }
            }
          }
        }
        """


@attrs.define(auto_attribs=True, kw_only=True)
class DbtCredentials(DbtCloudBase):
    account_id: str
    token: str = attrs.field(repr=lambda value: "***")

    def can_connect(self) -> bool:
        url = f"{self.base_v2}/accounts/"
        response = requests.get(url, headers=self.bearer_token_header)
        return response.status_code == 200

    @property
    def bearer_token_header(self) -> dict:
        return {"Authorization": f"Bearer {self.token}"}


@attrs.define(auto_attribs=True, kw_only=True)
class DbtExecutionInfo:
    last_success_run_id: str
    execution_time: float

    @classmethod
    def from_dict(cls, data: dict) -> "DbtExecutionInfo":
        return structure(data, cls)


@attrs.define(auto_attribs=True, kw_only=True)
class DbtExecutionHistoryInfo:
    run_id: str
    execution_time: float
    status: str

    @classmethod
    def from_dict(cls, data: dict) -> "DbtExecutionHistoryInfo":
        return structure(data, cls)


@attrs.define(auto_attribs=True, kw_only=True)
class DbtExecution:
    name: str
    unique_id: str
    execution_info: DbtExecutionInfo

    @classmethod
    def from_dict(cls, data: dict) -> "DbtExecution":
        return structure(data, cls)


@attrs.define(auto_attribs=True, kw_only=True)
class DbtConnection(DbtCloudBase):
    credentials: DbtCredentials

    def __attrs_post_init__(self) -> None:
        can_connect = self.credentials.can_connect()
        if not can_connect:
            raise ValueError("Cannot connect to dbt Cloud. Please, check credentials.")

    def _post_rest(self, url: str, data: dict) -> dict:
        response = requests.post(url, headers=self.credentials.bearer_token_header, data=data)
        if response.status_code != 200:
            raise ValueError(
                f"Response code {response.status_code} was returned. With the following message {response.text}."
            )
        return response.json()

    def _post_graphql(self, query: str, variables: dict) -> dict:
        response = requests.post(
            self.base_graphql,
            headers=self.credentials.bearer_token_header,
            json={"query": query, "variables": variables},
        )
        if response.status_code != 200:
            raise ValueError(
                f"Response code {response.status_code} was returned. With the following message {response.text}."
            )
        return response.json()

    def run_job(self, logger: logging.Logger, job_id: str) -> tuple[str, str]:
        url = f"{self.base_v2}/accounts/{self.credentials.account_id}/jobs/{job_id}/run/"
        data = {"cause": "Triggered via API by gooddata-dbt plugin"}
        # Allow testing from localhost where COMMIT_SHA is not set
        # GitHub does not propagate CI_COMMIT_SHA to environment variable in workers
        # Developers must propagate it manually from ${{ github.event.pull_request.head.sha }} to GOODDATA_GITHUB_SHA
        # DOC: https://docs.github.com/en/actions/using-workflows/events-that-trigger-workflows#pull_request
        commit_sha = os.getenv("GOODDATA_GITHUB_SHA") or os.getenv("CI_COMMIT_SHA")
        if commit_sha:
            data = {
                "cause": f"Triggered via API by gooddata-dbt plugin - {commit_sha=}",
                "git_sha": commit_sha,
            }
        logger.info(f"Run dbt cloud job {job_id=} with cause={data['cause']}")
        result = self._post_rest(url, data)
        run_id = safeget(result, ["data", "id"])
        run_href = safeget(result, ["data", "href"])
        if run_id is None:
            raise ValueError("Could not find wanted content.")
        return run_id, run_href

    @staticmethod
    def _was_fetch_success(response: dict) -> bool:
        is_complete = safeget(response, ["data", "is_complete"])
        is_success = safeget(response, ["data", "is_success"])
        return is_complete and is_success

    @staticmethod
    def _is_fetch_done(response: dict) -> bool:
        is_complete = safeget(response, ["data", "is_complete"])
        is_error = safeget(response, ["data", "is_error"])
        if is_error:
            status_humanized = safeget(response, ["data", "status_humanized"])
            raise Exception(f"dbt cloud job failed {status_humanized=}")
        return is_complete

    def _call_get_dbt_endpoint(self, url: str) -> Response:
        response = requests.get(url, headers=self.credentials.bearer_token_header)
        if response.status_code != 200:
            raise ValueError(
                f"Response code {response.status_code} was returned. With the following message {response.text}."
            )
        return response

    def fetch_run_result(
        self,
        run_id: str,
        # TODO - allow users to configure this from outside
        timeout: float = 600.0,
        retry: float = 0.2,
        max_retry: float = 5.0,
    ) -> None:
        time_sum = 0.0
        url = f"{self.base_v2}/accounts/{self.credentials.account_id}/runs/{run_id}"
        while True:
            response = self._call_get_dbt_endpoint(url)
            result = response.json()
            if self._is_fetch_done(result) and time_sum <= timeout:
                break
            time.sleep(retry)
            time_sum += retry
            retry = min(retry * 2, max_retry)
        if not self._was_fetch_success(result):
            raise ValueError(f"Job with {run_id=} failed.")
        if time_sum >= timeout:
            raise ValueError(f"Job with {run_id=} timed out. The time out was {timeout}.")

    def download_artifact(self, run_id: str, identifier: str, path: Union[str, Path]) -> None:
        """
        :param run_id: Identifier of the run.
        :param identifier: E.g., manifest.json
        :param path: Path where to store file.
        """
        path = path if isinstance(path, Path) else Path(path)
        url = f"{self.base_v2}/accounts/{self.credentials.account_id}/runs/{run_id}/artifacts/{identifier}"
        response = self._call_get_dbt_endpoint(url)
        result = response.json()
        if not os.path.isdir(path):
            os.makedirs(path)
        with open(path / identifier, "w") as f:
            json.dump(result, f)

    def download_manifest(self, run_id: str, path: Union[str, Path] = Path("target")) -> None:
        self.download_artifact(run_id, "manifest.json", path)

    def read_env_vars(self, project_id: str, job_id: str) -> dict:
        url = (
            f"{self.base_v3}/accounts/{self.credentials.account_id}/projects/{project_id}"
            f"/environment-variables/job/?job_definition_id={job_id}"
        )
        response = self._call_get_dbt_endpoint(url)
        result = response.json()
        data = result["data"]
        env_vars = {}
        for k, v in data.items():
            env_vars[k] = safeget(v, ["project", "value"])
        return env_vars

    def get_environment_id(self, job_id: str) -> str:
        url = f"{self.base_v2}/accounts/{self.credentials.account_id}/jobs/{job_id}/"
        response = self._call_get_dbt_endpoint(url)
        result = response.json()
        environment_id = safeget(result, ["data", "environment_id"])
        return environment_id

    # TODO - now it supports only Snowflake, add all supported data sources
    def read_environment(
        self,
        environment_id: str,
        project_id: str,
    ) -> DbtCloudEnvironment:
        url = (
            f"{self.base_v3}/accounts/{self.credentials.account_id}/projects/{project_id}/"
            + f"environments/{environment_id}"
        )
        response = self._call_get_dbt_endpoint(url)
        result = response.json()
        return DbtCloudEnvironment.from_dict(result["data"])

    @staticmethod
    def make_profiles(
        environment: DbtCloudEnvironment,
        path_to_profiles: Path,
        profile_file: str = "profiles.yml",
    ) -> None:
        if not os.path.isdir(path_to_profiles):
            os.makedirs(path_to_profiles)
        profile_file_path = path_to_profiles / profile_file
        profile = os.getenv("ELT_ENVIRONMENT")
        target = os.getenv("DBT_TARGET")
        db_pass = os.getenv("DBT_DB_PASS")
        schema_name = os.getenv("DBT_OUTPUT_SCHEMA")
        if not profile or not target or not db_pass or not schema_name:
            raise ValueError("Env variables ELT_ENVIRONMENT, DBT_TARGET, DBT_DB_PASS, DBT_OUTPUT_SCHEMA must be set.")
        data = {profile: {"outputs": {target: environment.to_profile(db_pass, schema_name)}}}

        with open(profile_file_path, "w", encoding="utf-8") as f:
            yaml.safe_dump(data, f, allow_unicode=True, sort_keys=False, encoding="utf-8")

    @staticmethod
    def string_camel_to_snake(element: str) -> str:
        return "".join(["_" + c.lower() if c.isupper() else c for c in element]).lstrip("_")

    def dict_camel_to_snake(self, data: Union[dict, list]) -> Union[dict, list]:
        if isinstance(data, list):
            result = []
            for record in data:
                if isinstance(record, dict):
                    result.append(self.dict_camel_to_snake(record))
                else:
                    return record
        else:
            result = {}  # type: ignore
            for key, value in data.items():
                if isinstance(value, dict):
                    result[self.string_camel_to_snake(key)] = self.dict_camel_to_snake(value)  # type: ignore
                else:
                    result[self.string_camel_to_snake(key)] = value  # type: ignore
        return result

    def get_last_execution(self, environment_id: str, model_count: int) -> list[DbtExecution]:
        variables = {"environmentId": environment_id, "first": model_count}
        result = self._post_graphql(self.graphql_applied_models, variables)
        model_edges = self.dict_camel_to_snake(safeget(result, ["data", "environment", "applied", "models", "edges"]))
        return [DbtExecution.from_dict(m["node"]) for m in model_edges]

    def get_average_times(
        self, logger: logging.Logger, models: list[DbtExecution], environment_id: str, history_count: int
    ) -> dict[str, float]:
        models_history_avg_execution_times = {}
        for model in models:
            variables = {"environmentId": environment_id, "modelId": model.unique_id, "first": history_count}
            logger.debug(f"Calling modelHistoricalRuns with {variables=}")
            try:
                result = self._post_graphql(self.graphql_model_historical_runs, variables)
            except Exception as e:
                logger.error(f"modelHistoricalRuns API returned error: {str(e)}")
                raise Exception("Error calling modelHistoricalRuns")
            model_historical_runs = self.dict_camel_to_snake(
                safeget(result, ["data", "environment", "applied", "modelHistoricalRuns"])
            )
            execs = [DbtExecutionHistoryInfo.from_dict(r) for r in model_historical_runs]
            exec_times = [e.execution_time for e in execs]
            avg_exec_time = sum(exec_times) / len(exec_times)
            models_history_avg_execution_times[model.unique_id] = avg_exec_time

        return models_history_avg_execution_times



# (C) 2023 GoodData Corporation
import asyncio
import logging
from argparse import Namespace
from asyncio import Semaphore
from pathlib import Path
from time import time
from typing import Optional

import tabulate
import yaml
from gooddata_sdk import CatalogDeclarativeModel, CatalogScanModelRequest, CatalogWorkspace, GoodDataSdk, Visualization

from gooddata_dbt.args import parse_arguments
from gooddata_dbt.dbt.cloud import DbtConnection, DbtCredentials, DbtExecution
from gooddata_dbt.dbt.profiles import DbtProfiles
from gooddata_dbt.dbt.tables import DbtModelTables
from gooddata_dbt.gooddata.api_wrapper import GoodDataApiWrapper
from gooddata_dbt.gooddata.config import GoodDataConfig, GoodDataConfigOrganization, GoodDataConfigProduct
from gooddata_dbt.logger import get_logger
from gooddata_dbt.sdk_wrapper import GoodDataSdkWrapper
from gooddata_dbt.utils import get_duration, report_message_to_git_vendor

# TODO - upgrade AIO, cleanup, start from scratch, test everything

GOODDATA_LAYOUTS_DIR = Path("gooddata_layouts")


def layout_model_path(data_product: GoodDataConfigProduct) -> Path:
    return GOODDATA_LAYOUTS_DIR / data_product.id


def generate_and_put_ldm(
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    data_source_id: str,
    workspace_id: str,
    dbt_tables: DbtModelTables,
    model_ids: Optional[list[str]],
) -> None:
    scan_request = CatalogScanModelRequest(scan_tables=True, scan_views=True)
    logger.info(f"Scan data source {data_source_id=}")
    scan_pdm = sdk_wrapper.sdk_facade.scan_data_source(data_source_id, scan_request)
    scan_pdm.store_to_disk(Path("test"))
    # Store data types to dbt_tables class. It is used in make_declarative_datasets to inject data types to LDM.
    dbt_tables.set_data_types(scan_pdm, sdk_wrapper.sdk_facade.dry_run)
    # Construct GoodData LDM from dbt models
    declarative_datasets = dbt_tables.make_declarative_datasets(data_source_id, model_ids)
    ldm = CatalogDeclarativeModel.from_dict({"ldm": declarative_datasets}, camel_case=False)
    # Deploy logical into target workspace
    sdk_wrapper.sdk_facade.put_declarative_ldm(workspace_id, ldm)


def create_workspace(
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    workspace_id: str,
    workspace_title: str,
) -> None:
    logger.info(f"Create workspace {workspace_id=} {workspace_title=}")
    # Create workspaces, if they do not exist yet, otherwise update them
    workspace = CatalogWorkspace(workspace_id=workspace_id, name=workspace_title)
    sdk_wrapper.sdk_facade.create_workspace(workspace=workspace)


def deploy_ldm(
    logger: logging.Logger,
    args: Namespace,
    all_model_ids: list[str],
    sdk_wrapper: GoodDataSdkWrapper,
    model_ids: Optional[list[str]],
    workspace_id: str,
) -> None:
    logger.info("Generate and put LDM")
    dbt_profiles = DbtProfiles(args)
    data_source_id = dbt_profiles.data_source_id
    dbt_tables = DbtModelTables.from_local(args.gooddata_upper_case, all_model_ids)
    generate_and_put_ldm(logger, sdk_wrapper, data_source_id, workspace_id, dbt_tables, model_ids)
    workspace_url = f"{sdk_wrapper.get_host_from_sdk()}/modeler/#/{workspace_id}"
    logger.info(f"LDM successfully loaded, verify here: {workspace_url}")


def register_data_source(
    logger: logging.Logger,
    args: Namespace,
    all_model_ids: list[str],
    sdk_wrapper: GoodDataSdkWrapper,
) -> None:
    dbt_profiles = DbtProfiles(args)
    dbt_target = dbt_profiles.target
    data_source_id = dbt_profiles.data_source_id
    logger.info(f"Process data source {data_source_id=}")
    dbt_tables = DbtModelTables.from_local(args.gooddata_upper_case, all_model_ids)

    logger.info(f"Register data source {data_source_id=} schema={dbt_tables.schema_name}")
    data_source = dbt_target.to_gooddata(data_source_id, dbt_tables.schema_name)
    sdk_wrapper.sdk_facade.create_or_update_data_source(data_source)


def upload_notification(logger: logging.Logger, sdk_wrapper: GoodDataSdkWrapper, data_source_id: str) -> None:
    logger.info(f"Upload notification {data_source_id=}")
    sdk_wrapper.sdk_facade.register_upload_notification(data_source_id)


def deploy_analytics(
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    workspace_id: str,
    data_product: GoodDataConfigProduct,
) -> None:
    logger.info(f"Deploy analytics {workspace_id=}")

    logger.info("Read analytics model from disk")
    adm = sdk_wrapper.sdk.catalog_workspace_content.load_analytics_model_from_disk(layout_model_path(data_product))

    # Deploy analytics model into target workspace
    logger.info("Load analytics model into GoodData")
    sdk_wrapper.sdk_facade.put_declarative_analytics_model(workspace_id, adm)

    workspace_url = f"{sdk_wrapper.get_host_from_sdk()}/dashboards/#/workspace/{workspace_id}"
    logger.info(f"Analytics successfully loaded, verify here: {workspace_url}")


def store_analytics(
    logger: logging.Logger, sdk: GoodDataSdk, workspace_id: str, data_product: GoodDataConfigProduct
) -> None:
    logger.info("Store analytics model to disk")
    sdk.catalog_workspace_content.store_analytics_model_to_disk(
        workspace_id,
        layout_model_path(data_product),
        # Exclude attributes related to activities of users, e.g. createdBy
        # When delivering programmatically,
        #   we don't want to transfer info about users and their activities into another environments
        exclude=["ACTIVITY_INFO"],
    )


async def execute_visualization(
    sdk_wrapper: GoodDataSdkWrapper, workspace_id: str, visualization: Visualization
) -> None:
    sdk_wrapper.sdk_facade.execute_visualization(workspace_id, visualization)


async def test_visualization(
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    workspace_id: str,
    visualization: Visualization,
) -> dict:
    logger.info(f"Executing visualization {visualization.id=} {visualization.title=} ...")
    start = time()
    try:
        await execute_visualization(sdk_wrapper, workspace_id, visualization)
        duration = get_duration(start)
        logger.info(f"Test successful {visualization.id=} {visualization.title=} duration={duration}(ms)")
        return {"id": visualization.id, "title": visualization.title, "duration": duration, "status": "success"}
    except Exception as e:
        duration = get_duration(start)
        logger.error(f"Test failed {visualization.id=} {visualization.title=} duration={duration}(ms) reason={str(e)}")
        return {
            "id": visualization.id,
            "title": visualization.title,
            "duration": duration,
            "status": "failed",
            "reason": str(e),
        }


async def safe_test_visualization(
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    workspace_id: str,
    visualization: Visualization,
    semaphore: Semaphore,
) -> dict:
    async with semaphore:  # semaphore limits num of simultaneous executions
        return await test_visualization(
            logger,
            sdk_wrapper,
            workspace_id,
            visualization,
        )


async def test_visualizations(
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    workspace_id: str,
    skip_tests: Optional[list[str]],
    test_visualizations_parallelism: int = 1,
) -> None:
    start = time()
    logger.info(f"Test visualizations {workspace_id=}")
    visualizations = sdk_wrapper.sdk_facade.get_visualizations(workspace_id)
    semaphore = asyncio.Semaphore(test_visualizations_parallelism)
    tasks = []
    for visualization in visualizations:
        if skip_tests is not None and visualization.id in skip_tests:
            logger.info(f"Skip test visualization={visualization.title} (requested in gooddata.yaml)")
        else:
            tasks.append(safe_test_visualization(logger, sdk_wrapper, workspace_id, visualization, semaphore))
    results = await asyncio.gather(*tasks)
    duration = get_duration(start)
    errors = [result for result in results if result["status"] == "failed"]
    if len(errors) > 0:
        raise Exception(f"Test visualizations failed {workspace_id=} {duration=}(ms) {errors=}")
    else:
        logger.info(f"Test visualizations finished {workspace_id=} {duration=}(ms)")


def create_localized_workspaces(
    logger: logging.Logger,
    data_product: GoodDataConfigProduct,
    sdk_facade: GoodDataApiWrapper,
    workspace_id: str,
) -> None:
    if data_product.localization is None:
        return
    try:
        from deep_translator import GoogleTranslator
    except ImportError:
        logger.warning(
            "create_localized_workspaces: deep_translator module not found, will not be able to localize workspaces"
        )
    for to in data_product.localization.to:
        translator_func = GoogleTranslator(
            source=data_product.localization.from_language, target=to.language
        ).translate_batch
        logger.info(f"create_localized_workspaces layout_root_path={GOODDATA_LAYOUTS_DIR / data_product.id}")
        sdk_facade.generate_localized_workspaces(
            workspace_id,
            to=to,
            data_product=data_product,
            translator_func=translator_func,
            layout_path=GOODDATA_LAYOUTS_DIR / data_product.id,
            provision_workspace=True,
            store_layouts=False,
        )


def get_table(data: list[list], headers: list[str], fmt: str) -> str:
    return tabulate.tabulate(data, headers=headers, tablefmt=fmt)


def dbt_cloud_stats_degradations(
    args: Namespace,
    logger: logging.Logger,
    environment_id: str,
    model_executions: list[DbtExecution],
    dbt_conn: DbtConnection,
) -> None:
    logger.info("Get stats for historical executions...")
    models_history_avg_execution_times = dbt_conn.get_average_times(logger, model_executions, environment_id, 5)
    differences = []
    degradations = 0
    for execution in model_executions:
        for model_id, avg_time in models_history_avg_execution_times.items():
            last_execution = execution.execution_info.execution_time
            difference = float(last_execution - avg_time) / avg_time * 100
            degradation = difference > args.allowed_degradation
            if model_id == execution.unique_id:
                differences.append(
                    [
                        model_id,
                        round(last_execution, 2),
                        round(avg_time, 2),
                        round(difference, 2),
                        str(degradation),
                    ]
                )
                if degradation:
                    degradations += 1
    headers = ["Model ID", "Last duration(s)", "Average duration(s)", "Difference(%)", "Degradation"]
    pretty_table = get_table(differences, headers, "outline")
    if degradations > 0:
        logger.warning(
            "Stats for historical executions contain degradations! "
            + f"Threshold={args.allowed_degradation}%\n{pretty_table}"
        )
    else:
        logger.info(f"Stats for historical executions:\n{pretty_table}")

    git_table = get_table(differences, headers, "github")

    report_message_to_git_vendor(logger, degradations, args.allowed_degradation, git_table)


def dbt_cloud_stats(
    args: Namespace,
    logger: logging.Logger,
    all_model_ids: list[str],
    environment_id: str,
) -> None:
    logger.info("Get stats for last execution...")
    dbt_conn = DbtConnection(credentials=DbtCredentials(account_id=args.account_id, token=args.token))
    dbt_tables = DbtModelTables.from_local(args.gooddata_upper_case, all_model_ids)
    model_executions = dbt_conn.get_last_execution(environment_id, len(dbt_tables.tables))
    exec_list = [
        [execution.unique_id, round(execution.execution_info.execution_time, 2)] for execution in model_executions
    ]

    headers = ["Model ID", "Duration(s)"]
    pretty_table = get_table(exec_list, headers, "outline")
    logger.info(f"Stats for last execution:\n{pretty_table}")

    dbt_cloud_stats_degradations(args, logger, environment_id, model_executions, dbt_conn)


def dbt_cloud_run(args: Namespace, logger: logging.Logger, all_model_ids: list[str]) -> None:
    dbt_conn = DbtConnection(credentials=DbtCredentials(account_id=args.account_id, token=args.token))
    logger.info("#" * 80)
    logger.info(f"Starting job in dbt cloud with job_id={args.job_id}")
    logger.info("#" * 80)

    run_id, run_href = dbt_conn.run_job(logger, args.job_id)
    logger.info(f"Job with {run_id=} successfully initiated. Link to job run {run_href=}")

    dbt_conn.fetch_run_result(run_id)
    dbt_conn.download_manifest(run_id)
    logger.info("Manifest downloaded successfully")

    environment_id = dbt_conn.get_environment_id(args.job_id)
    logger.info(f"Found environment with {environment_id=} for job_id={args.job_id}")

    environment = dbt_conn.read_environment(environment_id, args.project_id)
    dbt_conn.make_profiles(environment, path_to_profiles=Path(args.profiles_dir))
    logger.info(f"Profile file stored to {args.profiles_dir}")

    dbt_cloud_stats(args, logger, all_model_ids, environment_id)


def process_organization(
    args: Namespace,
    logger: logging.Logger,
    sdk_wrapper: GoodDataSdkWrapper,
    gd_config: GoodDataConfig,
    organization: Optional[GoodDataConfigOrganization] = None,
) -> None:
    if args.method == "upload_notification":
        dbt_profiles = DbtProfiles(args)
        # Caches are invalidated only per data source, not per data product
        upload_notification(logger, sdk_wrapper, dbt_profiles.data_source_id)
    elif args.method == "register_data_sources":
        register_data_source(logger, args, gd_config.all_model_ids, sdk_wrapper)
    else:
        if organization:
            data_products = [dp for dp in gd_config.data_products if dp.id in organization.data_product_ids]
        else:
            data_products = gd_config.data_products
        for data_product in data_products:
            logger.info(f"Process product name={data_product.name}")
            environments = gd_config.get_environment_workspaces(data_product.environment_setup_id)
            for environment in environments:
                if environment.id == args.gooddata_environment_id:
                    workspace_id = f"{data_product.id}_{environment.id}"
                    workspace_title = f"{data_product.name} ({environment.name})"
                    if args.method == "provision_workspaces":
                        create_workspace(logger, sdk_wrapper, workspace_id, workspace_title)
                    elif args.method == "deploy_ldm":
                        deploy_ldm(
                            logger, args, gd_config.all_model_ids, sdk_wrapper, data_product.model_ids, workspace_id
                        )
                    elif args.method == "store_analytics":
                        store_analytics(logger, sdk_wrapper.sdk, workspace_id, data_product)
                    elif args.method == "deploy_analytics":
                        deploy_analytics(logger, sdk_wrapper, workspace_id, data_product)
                        if data_product.localization:
                            create_localized_workspaces(logger, data_product, sdk_wrapper.sdk_facade, workspace_id)
                    elif args.method == "test_visualizations":
                        parallelism = gd_config.global_properties.test_visualizations_parallelism or 1
                        asyncio.run(
                            test_visualizations(logger, sdk_wrapper, workspace_id, data_product.skip_tests, parallelism)
                        )
                    else:
                        raise Exception(f"Unsupported method requested in args: {args.method}")


def main() -> None:
    args = parse_arguments("gooddata-dbt plugin for models management and invalidating caches(upload notification)")
    logger = get_logger("gooddata-dbt", args.debug)
    logger.info("Start")
    with open(args.gooddata_config) as fp:
        gd_config = GoodDataConfig.from_dict(yaml.safe_load(fp))
    # dbt cloud tasks must run only once, not for each organization
    if args.method == "dbt_cloud_run":
        dbt_cloud_run(args, logger, gd_config.all_model_ids)
    elif args.method == "dbt_cloud_stats":
        dbt_cloud_stats(args, logger, gd_config.all_model_ids, args.environment_id)
    else:
        if args.gooddata_profiles:
            logger.info(f"Process multiple organizations profiles={args.gooddata_profiles}")
            for organization in gd_config.organizations:
                if organization.gooddata_profile in args.gooddata_profiles:
                    sdk_wrapper = GoodDataSdkWrapper(args, logger, profile=organization.gooddata_profile)
                    logger.info(f"Process organization profile={organization.gooddata_profile}")
                    process_organization(args, logger, sdk_wrapper, gd_config, organization)
        else:
            sdk_wrapper = GoodDataSdkWrapper(args, logger)
            logger.info(f"Process single organization from env vars host={args.gooddata_host}")
            process_organization(args, logger, sdk_wrapper, gd_config)

    logger.info("End")



# (C) 2023 GoodData Corporation
import argparse
import os
import re
from typing import Optional, Union
from urllib.parse import quote_plus

import attrs
import yaml
from gooddata_sdk import (
    BasicCredentials,
    CatalogDataSourceMotherDuck,
    CatalogDataSourcePostgres,
    CatalogDataSourceRedshift,
    CatalogDataSourceSnowflake,
    CatalogDataSourceVertica,
    MotherDuckAttributes,
    PostgresAttributes,
    RedshiftAttributes,
    SnowflakeAttributes,
    TokenCredentialsFromEnvVar,
    VerticaAttributes,
)

from gooddata_dbt.dbt.base import Base


@attrs.define(auto_attribs=True, kw_only=True)
class DbtOutputPostgreSQL(Base):
    name: str
    title: str
    host: str
    port: str
    user: str
    password: str = attrs.field(repr=lambda value: "***")
    dbname: str
    database: str = attrs.field(default=attrs.Factory(lambda self: self.dbname, takes_self=True))
    schema: str

    def to_gooddata(self, data_source_id: str, schema_name: str) -> CatalogDataSourcePostgres:
        return CatalogDataSourcePostgres(
            id=data_source_id,
            name=self.title,
            db_specific_attributes=PostgresAttributes(
                host=self.host,
                port=self.port,
                # TODO - adopt this in Python SDK
                db_name=quote_plus(self.dbname),
            ),
            # Schema name is collected from dbt manifest from relevant tables
            schema=schema_name,
            credentials=BasicCredentials(
                username=self.user,
                password=self.password,
            ),
        )


@attrs.define(auto_attribs=True, kw_only=True)
class DbtOutputRedshift(Base):
    name: str
    title: str
    host: str
    port: str
    user: str
    password: str = attrs.field(repr=lambda value: "***")
    dbname: str
    database: str = attrs.field(default=attrs.Factory(lambda self: self.dbname, takes_self=True))
    schema: str

    def to_gooddata(self, data_source_id: str, schema_name: str) -> CatalogDataSourceRedshift:
        return CatalogDataSourceRedshift(
            id=data_source_id,
            name=self.title,
            db_specific_attributes=RedshiftAttributes(
                host=self.host,
                port=self.port,
                # TODO - adopt this in Python SDK
                db_name=quote_plus(self.dbname),
            ),
            # Schema name is collected from dbt manifest from relevant tables
            schema=schema_name,
            credentials=BasicCredentials(
                username=self.user,
                password=self.password,
            ),
        )


@attrs.define(auto_attribs=True, kw_only=True)
class DbtOutputSnowflake(Base):
    name: str
    title: str
    account: str
    user: str
    password: str = attrs.field(repr=lambda value: "***")
    database: str
    warehouse: str
    schema: str
    query_tag: Optional[str] = None

    def to_gooddata(self, data_source_id: str, schema_name: str) -> CatalogDataSourceSnowflake:
        return CatalogDataSourceSnowflake(
            id=data_source_id,
            name=self.title,
            db_specific_attributes=SnowflakeAttributes(
                # TODO - adopt this in Python SDK
                db_name=quote_plus(self.database),
                account=self.account,
                warehouse=self.warehouse,
            ),
            # Schema name is collected from dbt manifest from relevant tables
            schema=schema_name,
            credentials=BasicCredentials(
                username=self.user,
                password=self.password,
            ),
        )


@attrs.define(auto_attribs=True, kw_only=True)
class DbtOutputVertica(Base):
    name: str
    title: str
    host: str
    port: str
    username: str
    password: str = attrs.field(repr=lambda value: "***")
    database: str
    schema: str

    def to_gooddata(self, data_source_id: str, schema_name: str) -> CatalogDataSourceVertica:
        return CatalogDataSourceVertica(
            id=data_source_id,
            name=self.title,
            db_specific_attributes=VerticaAttributes(
                host=self.host,
                port=self.port,
                # TODO - adopt this in Python SDK
                db_name=quote_plus(self.database),
            ),
            # Schema name is collected from dbt manifest from relevant tables
            schema=schema_name,
            credentials=BasicCredentials(
                username=self.username,
                password=self.password,
            ),
        )


@attrs.define(auto_attribs=True, kw_only=True)
class DbtOutputMotherDuck(Base):
    name: str
    title: str
    path: str
    schema: str
    database: str = ""

    def __attrs_post_init__(self) -> None:
        self.database = self.validate_connection_props()

    def validate_connection_props(self) -> str:
        if not self.path.startswith("md:"):
            raise ValueError(f"Path {self.path} is not a valid MotherDuck path.")
        _, db_name = self.path.split(":", 1)
        if db_name:
            return db_name
        elif self.database:
            return self.database
        else:
            raise ValueError(f"Database name is neither specified in {self.path=} nor in {self.database=}")

    def to_gooddata(self, data_source_id: str, schema_name: str) -> CatalogDataSourceMotherDuck:
        return CatalogDataSourceMotherDuck(
            id=data_source_id,
            name=self.title,
            db_specific_attributes=MotherDuckAttributes(
                db_name=quote_plus(self.database),
            ),
            # Schema name is collected from dbt manifest from relevant tables
            schema=schema_name,
            credentials=TokenCredentialsFromEnvVar(
                env_var_name="MOTHERDUCK_TOKEN",
            ),
        )


DbtOutput = Union[DbtOutputPostgreSQL, DbtOutputRedshift, DbtOutputSnowflake, DbtOutputVertica, DbtOutputMotherDuck]


@attrs.define(auto_attribs=True, kw_only=True)
class DbtProfile(Base):
    name: str
    outputs: list[DbtOutput]


class DbtProfiles:
    """
    TODO:
        * from class methods from_cloud, from_local/core
        * How to read password for dbt Cloud?
    """

    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args
        with open(os.path.expanduser(f"{args.profiles_dir}/profiles.yml")) as fp:
            self.dbt_profiles = yaml.safe_load(fp)

    @staticmethod
    def inject_env_vars(output_def: dict) -> None:
        env_re = re.compile(r"\{\{ env_var\('([^']+)'(,\s*'([^']+)')?\) \}\}")
        for output_key, output_value in output_def.items():
            if (env_match := env_re.search(str(output_value))) is not None:
                default_value = None
                if len(env_match.groups()) == 3:
                    default_value = env_match.group(3)
                final_value = os.getenv(env_match.group(1)) or default_value
                if final_value is None:
                    output_def[output_key] = ""
                else:
                    output_def[output_key] = env_re.sub(final_value, str(output_value))
            # else do nothing, real value seems to be stored in dbt profile

    @staticmethod
    def to_data_class(output: str, output_def: dict) -> Optional[DbtOutput]:
        db_type = output_def["type"]
        if db_type == "postgres":
            return DbtOutputPostgreSQL.from_dict({"name": output, **output_def})
        elif db_type == "redshift":
            return DbtOutputRedshift.from_dict({"name": output, **output_def})
        elif db_type == "snowflake":
            return DbtOutputSnowflake.from_dict({"name": output, **output_def})
        elif db_type == "vertica":
            return DbtOutputVertica.from_dict({"name": output, **output_def})
        elif db_type == "duckdb" and output_def["path"].startswith("md:"):
            return DbtOutputMotherDuck.from_dict({"name": output, **output_def})
        elif db_type == "duckdb":
            # No logging available here. Pass because GoodData cannot connect to DuckDB file.
            return None
        else:
            raise Exception(f"Unsupported database type {output=} {db_type=}")

    @property
    def profiles(self) -> list[DbtProfile]:
        profiles = []
        for profile, profile_def in self.dbt_profiles.items():
            outputs = []
            for output, output_def in profile_def["outputs"].items():
                self.inject_env_vars(output_def)
                dbt_output = self.to_data_class(output, output_def)
                if dbt_output:
                    outputs.append(dbt_output)
            profiles.append(DbtProfile(name=profile, outputs=outputs))
        return profiles

    @property
    def profile(self) -> DbtProfile:
        for profile in self.profiles:
            if profile.name == self.args.profile:
                return profile
        raise ValueError(f"Profile {self.args.profile} not found in {self.profiles}.")

    @property
    def target(self) -> DbtOutput:
        for output in self.profile.outputs:
            if output.name == self.args.target:
                if self.args.gooddata_upper_case:
                    output.schema = output.schema.upper()
                    output.database = output.database.upper()
                return output
        raise ValueError(f"Target {self.args.target} not found in {self.profile.outputs}.")

    @property
    def data_source_id(self) -> str:
        return f"{self.args.profile}-{self.target.name}"



# (C) 2023 GoodData Corporation
import logging
from pathlib import Path
from typing import Any, Union

from gooddata_sdk import (
    CatalogDataSourceMotherDuck,
    CatalogDataSourcePostgres,
    CatalogDataSourceSnowflake,
    CatalogDataSourceVertica,
    CatalogDeclarativeAnalytics,
    CatalogDeclarativeColumn,
    CatalogDeclarativeModel,
    CatalogDeclarativeTable,
    CatalogDeclarativeTables,
    CatalogScanModelRequest,
    CatalogWorkspace,
    GoodDataSdk,
    Visualization,
)

from gooddata_dbt.gooddata.config import GoodDataConfigLocalizationTo, GoodDataConfigProduct

DataSource = Union[
    CatalogDataSourcePostgres,
    CatalogDataSourceSnowflake,
    CatalogDataSourceVertica,
    CatalogDataSourceMotherDuck,
]


class GoodDataApiWrapper:
    def __init__(self, sdk: GoodDataSdk, logger: logging.Logger, dry_run: bool = False) -> None:
        self.sdk = sdk
        self.logger = logger
        self.dry_run = dry_run

    def get_visualizations(self, workspace_id: str) -> list[Visualization]:
        if self.dry_run:
            self.logger.info("Dry run - skipping visualizations listing")
            return []
        else:
            return self.sdk.visualizations.get_visualizations(workspace_id)

    def execute_visualization(self, workspace_id: str, visualization: Visualization) -> None:
        if self.dry_run:
            self.logger.info("Dry run - skipping visualization execution")
        else:
            self.sdk.tables.for_visualization(workspace_id, visualization)

    def scan_data_source(self, data_source_id: str, scan_request: CatalogScanModelRequest) -> CatalogDeclarativeTables:
        if self.dry_run:
            self.logger.info("Dry run - skipping data source scanning")
            return CatalogDeclarativeTables(
                tables=[
                    CatalogDeclarativeTable(
                        id="dry_run",
                        type="DATA_SOURCE_TABLE",
                        path=["table"],
                        columns=[CatalogDeclarativeColumn(name="dry_run", data_type="STRING")],
                    )
                ]
            )
        else:
            return self.sdk.catalog_data_source.scan_data_source(data_source_id, scan_request, report_warnings=True).pdm

    def put_declarative_ldm(self, workspace_id: str, declarative_ldm: CatalogDeclarativeModel) -> None:
        if self.dry_run:
            self.logger.info("Dry run - skipping declarative LDM put")
        else:
            self.sdk.catalog_workspace_content.put_declarative_ldm(workspace_id, declarative_ldm)

    def create_workspace(self, workspace: CatalogWorkspace) -> None:
        if self.dry_run:
            self.logger.info("Dry run - skipping workspace creation")
        else:
            self.sdk.catalog_workspace.create_or_update(workspace=workspace)

    def create_or_update_data_source(self, data_source: DataSource) -> None:
        if self.dry_run:
            self.logger.info("Dry run - skipping data source creation")
        else:
            self.sdk.catalog_data_source.create_or_update_data_source(data_source)

    def register_upload_notification(self, data_source_id: str) -> None:
        if self.dry_run:
            self.logger.info("Dry run - skipping upload notification registration")
        else:
            self.sdk.catalog_data_source.register_upload_notification(data_source_id)

    def put_declarative_analytics_model(self, workspace_id: str, adm: CatalogDeclarativeAnalytics) -> None:
        if self.dry_run:
            self.logger.info("Dry run - skipping declarative analytics model put")
        else:
            self.sdk.catalog_workspace_content.put_declarative_analytics_model(workspace_id, adm)

    def generate_localized_workspaces(
        self,
        workspace_id: str,
        to: GoodDataConfigLocalizationTo,
        data_product: GoodDataConfigProduct,
        translator_func: Any,
        layout_path: Path,
        provision_workspace: bool = True,
        store_layouts: bool = False,
    ) -> None:
        from_language = "en"
        if data_product.localization is not None:
            from_language = data_product.localization.from_language
        if self.dry_run:
            self.logger.info("Dry run - skipping localized workspaces generation")
        else:
            self.sdk.catalog_workspace.generate_localized_workspaces(
                workspace_id,
                to_lang=to.language,
                to_locale=to.locale,
                from_lang=from_language,
                translator_func=translator_func,
                layout_root_path=layout_path,
                provision_workspace=provision_workspace,
                store_layouts=store_layouts,
            )



# (C) 2023 GoodData Corporation
from typing import Optional

import attr
import attrs

from gooddata_dbt.dbt.base import Base


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfigEnvironment(Base):
    id: str
    name: str


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfigEnvironmentSetup(Base):
    id: str
    environments: list[GoodDataConfigEnvironment]


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfigLocalizationTo(Base):
    locale: str
    language: str


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfigLocalization(Base):
    from_language: str
    to: list[GoodDataConfigLocalizationTo]


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfigProduct(Base):
    id: str
    name: str
    environment_setup_id: str
    model_ids: list[str] = attr.field(default=list)
    localization: Optional[GoodDataConfigLocalization] = None
    skip_tests: Optional[list[str]] = None


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfigOrganization(Base):
    gooddata_profile: str
    data_product_ids: list[str] = attr.field(default=list)


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataGlobalConfig(Base):
    test_visualizations_parallelism: Optional[int] = 1


@attrs.define(auto_attribs=True, kw_only=True)
class GoodDataConfig(Base):
    environment_setups: list[GoodDataConfigEnvironmentSetup]
    data_products: list[GoodDataConfigProduct]
    organizations: list[GoodDataConfigOrganization]
    global_properties: GoodDataGlobalConfig

    @property
    def all_model_ids(self) -> list[str]:
        result: list[str] = []
        for product in self.data_products:
            result.extend(product.model_ids)
        return result

    def get_environment_workspaces(self, environment_setup_id: str) -> list[GoodDataConfigEnvironment]:
        for env_setup in self.environment_setups:
            if env_setup.id == environment_setup_id:
                return env_setup.environments
        return []



# (C) 2023 GoodData Corporation

import attrs
from cattrs import structure


@attrs.define
class DbtCloudEnvironmentProjectConnectionDetails:
    account: str
    # database: str # TODO - dbt CLoud provides wrong value here, we must take it from Credentials, see below
    warehouse: str


@attrs.define
class DbtCloudEnvironmentProjectConnection:
    name: str
    type: str
    details: DbtCloudEnvironmentProjectConnectionDetails


@attrs.define
class DbtCloudEnvironmentProject:
    connection: DbtCloudEnvironmentProjectConnection


@attrs.define
class DbtCloudEnvironmentCredentials:
    user: str
    database: str


@attrs.define
class DbtCloudEnvironment:
    name: str
    project: DbtCloudEnvironmentProject
    credentials: DbtCloudEnvironmentCredentials

    @classmethod
    def from_dict(cls, data: dict) -> "DbtCloudEnvironment":
        return structure(data, cls)

    def to_profile(self, password: str, schema_name: str) -> dict:
        # TODO: add support for other DB types, now it is hardcoded to Snowflake
        return {
            "title": f"{self.project.connection.name} ({self.name} dbtCloud)",
            "type": self.project.connection.type,
            "account": self.project.connection.details.account,
            "user": self.credentials.user,
            "password": password,
            "database": self.credentials.database,
            "warehouse": self.project.connection.details.warehouse,
            "schema": schema_name,
        }



# (C) 2023 GoodData Corporation
import os
import re
import time
from logging import Logger

import requests


def post_gitlab_rest(url: str, headers: dict, data: dict) -> dict:
    response = requests.post(url, headers=headers, json=data)
    if response.status_code != 201:
        raise ValueError(
            f"Response code {response.status_code} was returned. With the following message {response.text}."
        )
    return response.json()


def report_message_to_merge_request(token: str, merge_request_id: str, text: str) -> None:
    url = (
        "https://gitlab.com/api/v4/projects/"
        + f"{os.getenv('CI_MERGE_REQUEST_PROJECT_ID')}/merge_requests/{merge_request_id}/notes"
    )
    headers = {"PRIVATE-TOKEN": token, "Content-Type": "application/json"}
    data = {"body": text}
    post_gitlab_rest(url, headers, data)


def report_message_to_pull_request(logger: Logger, token: str, pull_request_id: int, text: str) -> None:
    try:
        from github import Auth, Github

        auth = Auth.Token(token)
        g = Github(auth=auth)
        repo = g.get_repo(os.getenv("GITHUB_REPOSITORY"))
        issue = repo.get_issue(number=pull_request_id)
        print(f"{issue.user.id=} {issue.user.login=} {issue.user.name=}")
        issue.create_comment(text)
    except ImportError:
        logger.warning("PyGithub module not found, will not be able to report performance to GitHub pull request")


def report_message_to_git_vendor(logger: Logger, degradations: int, allowed_degradation: int, git_table: str) -> None:
    # Running in CI pipeline, report performance of executions to the merge/pull request to notify code reviewer
    gitlab_token = os.getenv("GITLAB_TOKEN")
    merge_request_id = os.getenv("CI_MERGE_REQUEST_IID")
    # Token/PR ID are exposed by GitHub actions in various ways depending on the event type
    # Let's decouple this code from it and expect GitHub workflows to set custom env variables
    github_token = os.getenv("GOODDATA_GITHUB_TOKEN")
    pr_id_str = os.getenv("GOODDATA_GITHUB_PULL_REQUEST_ID")
    pull_request_id = None
    if pr_id_str is not None and pr_id_str != "":
        if re.search(r"^[0-9]+$", pr_id_str) is not None:
            pull_request_id = int(pr_id_str)
        else:
            logger.warning(
                f"Pull request ID '{pr_id_str}' is not a number, "
                + "will not be able to report performance to GitHub pull request"
            )

    # Mention actor in GitHub comment to notify him. E-mail notifications are not sent to GitHub actors by default.
    github_actor = os.getenv("GOODDATA_GITHUB_ACTOR")
    prefix = ""
    if github_actor:
        prefix = f"@{github_actor} "
    if degradations > 0:
        message = (
            f"{prefix}**WARNING:** some executions of dbt models in this merge request are slower than average! "
            + f"Threshold = {allowed_degradation}%\n\n{git_table}"
        )
    else:
        message = f"{prefix}INFO: performance of all executions of dbt models is OK!\n\n{git_table}"

    if github_token and pull_request_id:
        logger.info(f"Sending report to related Github pull request as comment {pull_request_id=}")
        report_message_to_pull_request(logger, github_token, pull_request_id, message)
    elif gitlab_token and merge_request_id:
        logger.info(f"Sending report to related Gitlab merge request as comment {merge_request_id=}")
        report_message_to_merge_request(gitlab_token, merge_request_id, message)


def get_duration(start: float) -> int:
    return int((time.time() - start) * 1000)



# (C) 2023 GoodData Corporation
import json
import re
from typing import Optional

import attrs
from gooddata_sdk import CatalogDeclarativeMetric, CatalogDeclarativeModel

from gooddata_dbt.dbt.base import DBT_PATH_TO_MANIFEST, Base
from gooddata_dbt.dbt.tables import DbtModelBase

DBT_TO_GD_CALC_METHODS = {
    "count_distinct": "COUNT",
    "sum": "SUM",
    "average": "AVG",
    "min": "MIN",
    "max": "MAX",
}

DBT_TO_GD_FILTER_OPERATORS = {
    "!=": "<>",
    "is": "=",
    "is not": "<>",
}


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetaGoodDataMetricProps(Base):
    model_id: Optional[str] = None
    format: Optional[str] = None


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetaGoodDataMetric(Base):
    gooddata: DbtModelMetaGoodDataMetricProps = attrs.field(factory=DbtModelMetaGoodDataMetricProps)


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetricFilter(Base):
    field: str
    operator: str
    value: str


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetric(DbtModelBase):
    label: str
    meta: DbtModelMetaGoodDataMetric
    model: str
    calculation_method: str
    expression: str
    filters: Optional[list[DbtModelMetricFilter]] = None


class DbtModelMetrics:
    def __init__(self, model_ids: Optional[list[str]], ldm: CatalogDeclarativeModel) -> None:
        self.model_ids = model_ids
        self.ldm = ldm
        with open(DBT_PATH_TO_MANIFEST) as fp:
            self.dbt_catalog = json.load(fp)

    @property
    def metrics(self) -> list[DbtModelMetric]:
        result = [DbtModelMetric.from_dict(metric_def) for metric_def in self.dbt_catalog["metrics"].values()]
        # Return only gooddata labelled tables marked by model_id (if requested in args)
        return [
            r
            for r in result
            if r.meta.gooddata is not None and (not self.model_ids or r.meta.gooddata.model_id in self.model_ids)
        ]

    @staticmethod
    def extract_table_from_model(model: str) -> str:
        re_model = re.compile(r"ref\('([^']+)'\)")
        if (model_match := re_model.search(model)) is not None:
            return model_match.group(1)
        else:
            raise Exception("dbt model not specified by ref('xxx'), such breaking change not supported")

    def get_entity_type(self, table_name: str, expression_entity: str) -> str:
        result = None
        expression_entity_cmp = expression_entity.lower()
        if self.ldm.ldm is None:
            raise Exception("LDM cannot be None.")
        for dataset in self.ldm.ldm.datasets:
            table_id = dataset.data_source_table_id
            if table_id is not None and table_id.id.lower() == table_name.lower():
                attributes = dataset.attributes if dataset.attributes else []
                facts = dataset.facts if dataset.facts else []
                for attribute in attributes:
                    if attribute.source_column.lower() == expression_entity_cmp:
                        result = "label"
                    for label in attribute.labels:
                        if label.source_column.lower() == expression_entity_cmp:
                            result = "label"
                for fact in facts:
                    if fact.source_column.lower() == expression_entity_cmp:
                        result = "fact"
        for date_dataset in self.ldm.ldm.date_instances:
            if date_dataset.id.lower() == expression_entity_cmp:
                result = "date"
        if result:
            return result
        else:
            raise Exception(f"Unsupported entity type {table_name=} {expression_entity=}")

    def make_entity_id(self, table_name: str, token: str) -> Optional[str]:
        entity_type = self.get_entity_type(table_name, token)
        if not entity_type:
            return None
        full_entity_name = token
        final_entity_type = entity_type
        if entity_type == "date":
            # TODO - not sure how date dims are handled in dbt models, cannot find it in DOC
            full_entity_name = f"{token}.day"
            final_entity_type = "label"
        return f" {{{final_entity_type}/{full_entity_name}}}"

    def resolve_entities_in_expression(self, expression: str, table_name: str) -> str:
        re_split = re.compile(r"\s+")
        tokens = re_split.split(expression)
        result_tokens = []
        for token in tokens:
            entity_id = self.make_entity_id(table_name, token)
            result_tokens.append(entity_id or token)
        return " ".join(result_tokens)

    def make_gooddata_filter(self, table_name: str, dbt_filters: Optional[list[DbtModelMetricFilter]] = None) -> str:
        # TODO - Quite naive implementation
        #    e.g. missing polishing of values (e.g. SQL vs MAQL enclosers)
        gd_maql_filters = []
        dbt_filters = dbt_filters if dbt_filters is not None else []
        for dbt_filter in dbt_filters:
            entity_id = self.make_entity_id(table_name, dbt_filter.field)
            operator = DBT_TO_GD_FILTER_OPERATORS.get(dbt_filter.operator, dbt_filter.operator)
            gd_maql_filters.append(f"{entity_id} {operator} {dbt_filter.value}")
        if gd_maql_filters:
            return " WHERE " + " AND ".join(gd_maql_filters)
        else:
            return ""

    def make_gooddata_metrics(self) -> list[CatalogDeclarativeMetric]:
        gd_metrics = []
        for dbt_metric in self.metrics:
            calculation_method = DBT_TO_GD_CALC_METHODS.get(dbt_metric.calculation_method)
            if calculation_method:
                table_name = self.extract_table_from_model(dbt_metric.model)
                expression = self.resolve_entities_in_expression(dbt_metric.expression, table_name)
                gd_filter = self.make_gooddata_filter(table_name, dbt_metric.filters)

                gd_maql = f"SELECT {calculation_method}({expression}){gd_filter}"
                metric_dict = {
                    "id": dbt_metric.name,
                    "title": dbt_metric.label,
                    "description": dbt_metric.description,
                    "content": {
                        "format": dbt_metric.meta.gooddata.format,
                        "maql": gd_maql,
                    },
                    "tags": dbt_metric.tags,
                }
                gd_metrics.append(CatalogDeclarativeMetric.from_dict(metric_dict))
            else:
                raise Exception(f"Unsupported calculation method: {dbt_metric.calculation_method}")
        return gd_metrics



# (C) 2023 GoodData Corporation
import argparse
import os


def get_parser(description: str) -> argparse.ArgumentParser:
    return argparse.ArgumentParser(
        conflict_handler="resolve", description=description, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )


def set_gooddata_endpoint_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-gc",
        "--gooddata-config",
        help="Path to GoodData config file containing definitions of environments",
        default=os.getenv("GOODDATA_CONFIG", "gooddata.yml"),
    )
    parser.add_argument(
        "-gh",
        "--gooddata-host",
        help="Hostname(DNS) where GoodData is running",
        default=os.getenv("GOODDATA_HOST", "http://localhost:3000"),
    )
    parser.add_argument(
        "-gt",
        "--gooddata-token",
        help="GoodData API token for authentication",
        default=os.getenv("GOODDATA_TOKEN", "YWRtaW46Ym9vdHN0cmFwOmFkbWluMTIz"),
    )
    parser.add_argument(
        "-go",
        "--gooddata-override-host",
        help="Override hostname, if necessary. "
        "When you connect to different hostname than where GoodData is running(proxies)",
        default=os.getenv("GOODDATA_OVERRIDE_HOST"),
    )
    # Alternative - use profile.yml file
    env_profiles = os.getenv("GOODDATA_PROFILES")
    env_profiles_list = env_profiles.split(" ") if env_profiles else []
    parser.add_argument(
        "-gp",
        "--gooddata-profiles",
        nargs="*",
        help="Profiles in profile.yml file. Overrides gooddata-host, gooddata-token and gooddata-override-host."
        + "You can use multiple profiles separated by space to deliver models/analytics to multiple organizations.",
        default=env_profiles_list,
    )


def set_environment_id_arg(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-gw",
        "--gooddata-environment-id",
        help="Environment where to deploy. Environments are configured in gooddata.yml.",
        default=os.getenv("GOODDATA_ENVIRONMENT_ID", "development"),
    )


def set_gooddata_upper_case_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-guc",
        "--gooddata-upper-case",
        help="Upper case all physical model entities (tables, columns). Valuable for Snowflake!",
        action="store_true",
        default=False,
    )


def set_gooddata_workspace_title_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-gwt", "--gooddata-workspace-title", help="Workspace title", default=os.getenv("GOODDATA_WORKSPACE_TITLE")
    )


def set_dbt_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-pd",
        "--profiles-dir",
        help="Directory where dbt profiles.yml is stored",
        default=os.getenv("DBT_PROFILES_DIR", "~/.dbt"),
    )
    parser.add_argument(
        "-p", "--profile", help="Name of profile from profiles.yml", default=os.getenv("DBT_PROFILE", "default")
    )
    parser.add_argument(
        "-t",
        "--target",
        help="dbt target/output. DB where dbt deploys. GoodData registers it as data source.",
        default=os.getenv("ELT_ENVIRONMENT", "cicd_dev_local"),
    )


def set_dbt_cloud_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-dca",
        "--account-id",
        help="dbt cloud account ID",
        default=os.getenv("DBT_ACCOUNT_ID"),
    )
    parser.add_argument(
        "-dcj",
        "--job-id",
        help="dbt cloud job ID",
        default=os.getenv("DBT_JOB_ID"),
    )
    parser.add_argument(
        "-dcp",
        "--project-id",
        help="dbt cloud project ID",
        default=os.getenv("DBT_PROJECT_ID"),
    )
    parser.add_argument(
        "-dct",
        "--token",
        help="dbt cloud token",
        default=os.getenv("DBT_TOKEN"),
    )
    parser.add_argument(
        "-dcd",
        "--allowed-degradation",
        type=int,
        help="How much performance of model executions can degrade before reported as warning, in percent",
        default=int(os.getenv("DBT_ALLOWED_DEGRADATION", "100")),
    )


def set_dbt_cloud_stats_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-dce",
        "--environment-id",
        help="dbt cloud environment ID",
        default=os.getenv("DBT_ENVIRONMENT_ID"),
    )


def parse_arguments(description: str) -> argparse.Namespace:
    parser = get_parser(description)
    parser.add_argument("--debug", action="store_true", default=False, help="Increase logging level to DEBUG")
    parser.add_argument("--dry-run", action="store_true", default=False, help="Do not call GoodData APIs")
    set_gooddata_endpoint_args(parser)

    subparsers = parser.add_subparsers(help="actions")

    dbt_cloud_run = subparsers.add_parser("dbt_cloud_run")
    set_dbt_cloud_args(dbt_cloud_run)
    set_dbt_args(dbt_cloud_run)
    set_gooddata_upper_case_args(dbt_cloud_run)
    dbt_cloud_run.set_defaults(method="dbt_cloud_run")

    dbt_cloud_stats = subparsers.add_parser("dbt_cloud_stats")
    set_dbt_cloud_args(dbt_cloud_stats)
    set_dbt_cloud_stats_args(dbt_cloud_stats)
    set_dbt_args(dbt_cloud_stats)
    set_gooddata_upper_case_args(dbt_cloud_stats)
    dbt_cloud_stats.set_defaults(method="dbt_cloud_stats")

    provision_workspaces = subparsers.add_parser("provision_workspaces")
    set_environment_id_arg(provision_workspaces)
    provision_workspaces.set_defaults(method="provision_workspaces")

    register_data_sources = subparsers.add_parser("register_data_sources")
    set_dbt_args(register_data_sources)
    set_environment_id_arg(register_data_sources)
    set_gooddata_upper_case_args(register_data_sources)
    register_data_sources.set_defaults(method="register_data_sources")

    deploy_ldm = subparsers.add_parser("deploy_ldm")
    set_dbt_args(deploy_ldm)
    set_environment_id_arg(deploy_ldm)
    set_gooddata_upper_case_args(deploy_ldm)
    deploy_ldm.set_defaults(method="deploy_ldm")

    upload_notification = subparsers.add_parser("upload_notification")
    set_dbt_args(upload_notification)
    set_gooddata_upper_case_args(upload_notification)
    upload_notification.set_defaults(method="upload_notification")

    deploy_analytics = subparsers.add_parser("deploy_analytics")
    set_environment_id_arg(deploy_analytics)
    # TODO - now it is no longer needed. Either delete it or utilize it to do not lower_case everything
    set_gooddata_upper_case_args(deploy_analytics)
    deploy_analytics.set_defaults(method="deploy_analytics")

    store_analytics = subparsers.add_parser("store_analytics")
    set_environment_id_arg(store_analytics)
    set_gooddata_upper_case_args(store_analytics)
    store_analytics.set_defaults(method="store_analytics")

    test_visualizations = subparsers.add_parser("test_visualizations")
    set_environment_id_arg(test_visualizations)
    test_visualizations.set_defaults(method="test_visualizations")

    return parser.parse_args()



# (C) 2023 GoodData Corporation
import argparse
from logging import Logger
from typing import Optional

from gooddata_sdk import GoodDataSdk

from gooddata_dbt.gooddata.api_wrapper import GoodDataApiWrapper


class GoodDataSdkWrapper:
    # Timeout=600 because supporting waiting for All-in-one image starts
    def __init__(
        self, args: argparse.Namespace, logger: Logger, profile: Optional[str] = None, timeout: int = 600
    ) -> None:
        self.args = args
        self.logger = logger
        self.timeout = timeout
        self.profile = profile
        self.sdk = self.create_sdk()
        self.sdk_facade = self.create_sdk_facade()
        if not self.args.dry_run:
            self.wait_for_gooddata_is_up(self.timeout)

    def get_host_from_sdk(self) -> Optional[str]:
        # TODO - make _hostname public in gooddata_sdk
        return self.sdk.client._hostname

    def create_sdk(self) -> GoodDataSdk:
        if self.profile:
            self.logger.info(f"Connecting to GoodData using profile={self.profile}")
            sdk = GoodDataSdk.create_from_profile(profile=self.profile)
            return sdk
        else:
            host = self.args.gooddata_host
            token = self.args.gooddata_token
            override_host = self.args.gooddata_override_host
            kwargs = {}
            if override_host:
                kwargs["Host"] = override_host
            masked_token = f"{len(token[:-4]) * '#'}{token[-4:]}"
            self.logger.info(f"Connecting to GoodData host={host} token={masked_token} override_host={override_host}")
            sdk = GoodDataSdk.create(host_=host, token_=token, **kwargs)
            return sdk

    def create_sdk_facade(self) -> GoodDataApiWrapper:
        return GoodDataApiWrapper(self.sdk, self.logger, self.args.dry_run)

    def wait_for_gooddata_is_up(self, timeout: int) -> None:
        # Wait for the GoodData.CN docker image to start up or prevent hiccups of cloud deployments
        # We have to take hostname from sdk.client, because it can also be collected from profiles.yml file
        host = self.get_host_from_sdk()
        self.logger.info(f"Waiting for {host} to be up")
        self.sdk.support.wait_till_available(timeout=timeout)
        self.logger.info(f"Host {host} is up")

    def pre_cache_visualizations(self, workspaces: Optional[list] = None) -> None:
        if not workspaces:
            workspaces = [w.id for w in self.sdk.catalog_workspace.list_workspaces()]
        for workspace_id in workspaces:
            visualizations = self.sdk.visualizations.get_visualizations(workspace_id)

            for visualization in visualizations:
                self.sdk.tables.for_visualization(workspace_id, visualization)



# (C) 2023 GoodData Corporation
from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

REQUIRES = [
    "gooddata-sdk~=1.39.0",
    "pyyaml>=6.0",
    "attrs>=21.4.0,<=24.2.0",
    "cattrs>=22.1.0,<=24.1.1",
    "requests~=2.32.0",
    "tabulate~=0.8.10",
]

setup(
    name="gooddata-dbt",
    description="dbt plugin for GoodData",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="1.39.0",
    author="GoodData",
    author_email="support@gooddata.com",
    license="MIT",
    license_file="LICENSE.txt",
    license_files=("LICENSE.txt",),
    install_requires=REQUIRES,
    packages=find_packages(exclude=["tests*"]),
    python_requires=">=3.9.0",
    scripts=[
        "bin/gooddata-dbt",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Database",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Typing :: Typed",
    ],
    keywords=[],
    include_package_data=True,
)



# (C) 2023 GoodData Corporation
import json
from pathlib import Path
from typing import Union

from gooddata_dbt.dbt.tables import DbtModelTables
from gooddata_sdk import CatalogDeclarativeModel, CatalogDeclarativeTables

_CURR_DIR = Path(__file__).parent
_MANIFEST_PATH = _CURR_DIR / "resources/dbt_target/manifest.json"
_PDM_PATH = _CURR_DIR / "resources/gooddata_layouts"
MODEL_ID = "github"


def _read_json(path: Union[str, Path]) -> dict:
    with open(path) as f:
        return json.load(f)


def test_load_tables():
    tables = DbtModelTables.from_local(upper_case=False, all_model_ids=[MODEL_ID], manifest_path=_MANIFEST_PATH)

    assert len(tables.tables) == 4
    assert len(tables.tables[0].columns) == 6

    assert tables.tables[0].name == "commits"
    assert tables.tables[0].columns["created_at"].name == "created_at"


def test_load_tables_upper_case():
    tables = DbtModelTables.from_local(upper_case=True, all_model_ids=[MODEL_ID], manifest_path=_MANIFEST_PATH)

    assert len(tables.tables) == 4
    assert len(tables.tables[0].columns) == 6

    assert tables.tables[0].name == "COMMITS"
    assert tables.tables[0].columns["created_at"].name == "CREATED_AT"


def test_make_ldm():
    tables = DbtModelTables.from_local(upper_case=False, all_model_ids=[MODEL_ID], manifest_path=_MANIFEST_PATH)
    scan_pdm = CatalogDeclarativeTables.load_from_disk(_PDM_PATH)
    tables.set_data_types(scan_pdm)
    data_source_id = "postgres"

    declarative_datasets = tables.make_declarative_datasets(data_source_id, [MODEL_ID])
    ldm = CatalogDeclarativeModel.from_dict({"ldm": declarative_datasets}, camel_case=False)

    assert len(ldm.ldm.datasets) == 4
    assert len(ldm.ldm.date_instances) == 4



# (C) 2023 GoodData Corporation
import copy
import json
import re
from pathlib import Path
from typing import Optional, Union

import attrs
from gooddata_sdk import CatalogDeclarativeColumn, CatalogDeclarativeTable, CatalogDeclarativeTables
from gooddata_sdk.utils import safeget

from gooddata_dbt.dbt.base import (
    DATE_GRANULARITIES,
    DATETIME_DATA_TYPES,
    DBT_PATH_TO_MANIFEST,
    DBT_TARGET_DIR,
    NUMERIC_DATA_TYPES,
    TIMESTAMP_DATA_TYPES,
    TIMESTAMP_GRANULARITIES,
    Base,
    GoodDataLabelType,
    GoodDataLdmType,
    GoodDataSortDirection,
)
from gooddata_dbt.dbt.cloud import DbtConnection


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetaGoodDataTableProps(Base):
    model_id: Optional[str] = None


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetaGoodDataColumnProps(Base):
    id: Optional[str] = None
    ldm_type: Optional[GoodDataLdmType] = None
    referenced_table: Optional[str] = None
    label_type: Optional[GoodDataLabelType] = None
    attribute_column: Optional[str] = None
    sort_column: Optional[str] = None
    sort_direction: Optional[GoodDataSortDirection] = None
    default_view: Optional[bool] = None

    @property
    def gooddata_ref_table_ldm_id(self) -> Optional[str]:
        if self.referenced_table:
            return self.referenced_table.lower()
        return None

    def upper_case_names(self) -> None:
        if self.referenced_table:
            self.referenced_table = self.referenced_table.upper()
        if self.attribute_column:
            self.attribute_column = self.attribute_column.upper()
        if self.sort_column:
            self.sort_column = self.sort_column.upper()


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetaGoodDataTable(Base):
    gooddata: DbtModelMetaGoodDataTableProps = attrs.field(factory=DbtModelMetaGoodDataTableProps)


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelMetaGoodDataColumn(Base):
    gooddata: DbtModelMetaGoodDataColumnProps = attrs.field(factory=DbtModelMetaGoodDataColumnProps)


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelBase(Base):
    name: str
    description: str
    tags: list[str]
    # If 2+ references point to the same table, the table plays multiple roles,
    # it must be generated as multiple datasets
    role_name: Optional[str] = None

    # TODO - duplicate of backend logic.
    # Solution: use result of generateLdm as a master template, and override based on dbt metadata only if necessary
    @staticmethod
    def beautify_title(name: str) -> str:
        re_invalid_chars = re.compile(r"[^a-z0-9 ]", re.I)
        re_spaces = re.compile(r"\s+")
        if name == name.upper():
            # Snowflake-like DBs expose upper-case by default
            name = name.lower()
        name = re_invalid_chars.sub(" ", name)
        name = re_spaces.sub(" ", name)
        return name.capitalize()

    # Title is not supported by dbt, we fill it here
    @property
    def title(self) -> str:
        return self.description or self.beautify_title(self.name)

    @property
    def gooddata_ldm_id(self) -> str:
        if self.role_name:
            return f"{self.name.lower()}_{self.role_name.lower()}"
        else:
            return self.name.lower()

    @property
    def gooddata_ldm_title(self) -> str:
        if self.role_name:
            return f"{self.title} ({self.role_name.lower()})"
        else:
            return self.title

    @property
    def gooddata_ldm_description(self) -> str:
        if self.role_name and self.description:
            return f"{self.description} ({self.role_name.lower()})"
        else:
            return self.description

    def upper_case_name(self) -> None:
        self.name = self.name.upper()


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelColumn(DbtModelBase):
    data_type: Optional[str]
    meta: DbtModelMetaGoodDataColumn = attrs.field(factory=DbtModelMetaGoodDataColumn)

    # Enable to override LDM ID for LDM entities derived from columns (attributes, ...)
    # TODO - does it work for PK/references in combination with roles?
    @property
    def ldm_id(self) -> str:
        return self.meta.gooddata.id or self.gooddata_ldm_id

    @property
    def ldm_type(self) -> Optional[str]:
        if self.meta.gooddata.ldm_type is None:
            return None
        else:
            return self.meta.gooddata.ldm_type.value

    @property
    def label_type(self) -> Optional[str]:
        if self.meta.gooddata.label_type is None:
            return None
        else:
            return self.meta.gooddata.label_type.value

    @property
    def sort_direction(self) -> Optional[str]:
        if self.meta.gooddata.sort_direction is None:
            return None
        else:
            return self.meta.gooddata.sort_direction.value

    def gooddata_is_fact(self) -> bool:
        return (self.meta.gooddata.ldm_type == GoodDataLdmType.FACT) or self.is_number()

    def gooddata_is_attribute(self) -> bool:
        valid_ldm_types = [GoodDataLdmType.ATTRIBUTE, GoodDataLdmType.PRIMARY_KEY]
        # Without GD metadata, attribute is default unless it is DATETIME/NUMBER(FLOAT) data type
        return self.meta.gooddata.ldm_type in valid_ldm_types or (
            self.meta.gooddata.ldm_type is None and not self.is_date() and not self.is_number()
        )

    def gooddata_is_label(self, attribute_column_name: str) -> bool:
        return (
            self.meta.gooddata.ldm_type == GoodDataLdmType.LABEL
            and attribute_column_name == self.meta.gooddata.attribute_column
        )

    def is_date(self) -> bool:
        if self.data_type is None:
            return False
        gooddata_date = self.meta.gooddata.ldm_type == GoodDataLdmType.DATE
        return gooddata_date or self.data_type.upper() in DATETIME_DATA_TYPES

    def is_number(self) -> bool:
        if self.data_type is None:
            return False
        return self.data_type.upper() in NUMERIC_DATA_TYPES

    def is_reference(self) -> bool:
        return self.meta.gooddata.ldm_type == GoodDataLdmType.REFERENCE


@attrs.define(auto_attribs=True, kw_only=True)
class DbtModelTable(DbtModelBase):
    schema: str
    columns: dict[str, DbtModelColumn]
    meta: DbtModelMetaGoodDataTable = attrs.field(factory=DbtModelMetaGoodDataTable)


class DbtModelTables:
    """
    TODO:
        * add 2 class method – from_cloud, from_local
        * DbtModelTables should accept tables only
        * check dbt Cloud if it gives data type
            * if yes get rid off LDM if possible and other fun
        * column_type – Optional if missing call  scan
    """

    def __init__(self, tables: list[DbtModelTable], upper_case: bool) -> None:
        self.upper_case = upper_case
        self.tables = tables

    @classmethod
    def from_cloud(
        cls,
        dbt_conn: DbtConnection,
        run_id: str,
        upper_case: bool,
        all_model_ids: list[str],
        path: Union[str, Path] = DBT_TARGET_DIR,
    ) -> "DbtModelTables":
        path = path if isinstance(path, Path) else Path(path)
        dbt_conn.download_manifest(run_id=run_id, path=path)
        with open(path / "manifest.json") as fp:
            dbt_catalog = json.load(fp)
        tables = cls.read_dbt_models(dbt_catalog, upper_case, all_model_ids)
        return cls(tables, upper_case)

    @classmethod
    def from_local(
        cls, upper_case: bool, all_model_ids: list[str], manifest_path: Union[str, Path] = DBT_PATH_TO_MANIFEST
    ) -> "DbtModelTables":
        with open(manifest_path) as fp:
            dbt_catalog = json.load(fp)
        tables = cls.read_dbt_models(dbt_catalog, upper_case, all_model_ids)
        return cls(tables, upper_case)

    @staticmethod
    def read_dbt_models(dbt_catalog: dict, upper_case: bool, all_model_ids: list[str]) -> list[DbtModelTable]:
        tables = []
        for model_def in dbt_catalog["nodes"].values():
            model_id = safeget(model_def, ["meta", "gooddata", "model_id"])
            if model_id in all_model_ids:
                tables.append(DbtModelTable.from_dict(model_def))

        if len(tables) == 0:
            raise Exception("No tables labelled by gooddata meta flag found in the data source")

        if upper_case:
            for table in tables:
                table.upper_case_name()
                for column in table.columns.values():
                    column.upper_case_name()
                    column.meta.gooddata.upper_case_names()
        return tables

    def set_data_types(self, scan_pdm: CatalogDeclarativeTables, dry_run: bool = False) -> None:
        if dry_run:
            for table in self.tables:
                for column in table.columns.values():
                    column.data_type = "STRING"
        else:
            for table in self.tables:
                scan_table = self.get_scan_table(scan_pdm, table.name)
                for column in table.columns.values():
                    # dbt does not provide data types in manifest.json
                    # get it from GoodData scan API
                    scan_column = self.get_scan_column(scan_table, column.name)
                    column.data_type = scan_column.data_type

    @property
    def schema_name(self) -> str:
        schemas = set([t.schema for t in self.tables if t.schema])
        if len(schemas) > 1:
            raise Exception(f"Unsupported feature: GoodData does not support multiple schemas - {schemas=}")
        elif len(schemas) < 1:
            raise Exception("No schema found")
        else:
            schema_name = next(iter(schemas))
            if self.upper_case:
                return schema_name.upper()
            else:
                return schema_name

    @staticmethod
    def get_scan_table(scan_pdm: CatalogDeclarativeTables, table_name: str) -> CatalogDeclarativeTable:
        for table in scan_pdm.tables:
            if table.id.lower() == table_name.lower():
                return table
        scan_tables = [s.id for s in scan_pdm.tables]
        raise Exception(f"get_scan_table table={table_name} not found in scan! {scan_tables=}")

    @staticmethod
    def get_scan_column(table: CatalogDeclarativeTable, column_name: str) -> CatalogDeclarativeColumn:
        for column in table.columns:
            if column.name.lower() == column_name.lower():
                return column
        scan_columns = [s.name for s in table.columns]
        raise Exception(f"get_scan_column table={table.id} column={column_name} not found in scan. {scan_columns=}")

    @staticmethod
    def get_ldm_title(column: DbtModelColumn) -> str:
        return column.description or column.name

    @staticmethod
    def is_primary_key(column: DbtModelColumn) -> bool:
        result = False
        # TODO - constraints are stored in special nodes
        # for test in column.tests:
        #     if DbtTests.PRIMARY_KEY.value in test:
        #         result = True
        if column.meta.gooddata.ldm_type == GoodDataLdmType.PRIMARY_KEY:
            result = True
        return result

    def make_grain(self, table: DbtModelTable) -> list[dict]:
        grain = [
            {"id": column.ldm_id, "type": "attribute"}
            for column in table.columns.values()
            if self.is_primary_key(column)
        ]
        return grain

    # TODO - constraints are stored in special nodes
    # @staticmethod
    # def get_dbt_foreign_key(column: DbtModelColumn) -> Optional[str]:
    #     referenced_object_id = None
    #     if column.meta.gooddata.ldm_type == GoodDataLdmType.REFERENCE:
    #         referenced_object_id = column.meta.gooddata.gooddata_ref_table_ldm_id
    #     else:
    #         for test in column.tests:
    #             if DbtTests.FOREIGN_KEY.value in test:
    #                 referenced_object_id = test[DbtTests.FOREIGN_KEY.value][DbtTests.FOREIGN_KEY_REF.value]
    #     return referenced_object_id

    @staticmethod
    def find_role_playing_tables(tables: list[DbtModelTable]) -> dict:
        return {
            referenced_object_id: columns
            for table in tables
            for column in table.columns.values()
            if column.is_reference()
            for referenced_table in [column.meta.gooddata.referenced_table]
            if referenced_table is not None
            for referenced_object_id, columns in {
                referenced_table: [
                    col.name
                    for col in table.columns.values()
                    if col.is_reference() and col.meta.gooddata.referenced_table == referenced_table
                ]
            }.items()
            if len(columns) > 1
        }

    def make_references(self, table: DbtModelTable, role_playing_tables: dict) -> list[dict]:
        references = []
        for column in table.columns.values():
            referenced_object_id = None
            if column.is_reference():
                referenced_object_id = column.meta.gooddata.gooddata_ref_table_ldm_id
                referenced_object_name = referenced_object_id
                if referenced_object_name is not None:
                    if self.upper_case:
                        referenced_object_name = referenced_object_name.upper()
                    if referenced_object_name in role_playing_tables:
                        referenced_object_id = f"{referenced_object_id}_{column.ldm_id}"
            elif column.is_date():
                referenced_object_id = column.ldm_id
            if referenced_object_id is not None:
                references.append(
                    {
                        "identifier": {"id": referenced_object_id, "type": "dataset"},
                        "multivalue": False,
                        "source_columns": [column.name],
                        "source_column_data_types": [column.data_type],
                    }
                )
        return references

    @staticmethod
    def make_facts(table: DbtModelTable) -> list[dict]:
        facts = [
            {
                "id": column.ldm_id,
                # TODO - all titles filled from dbt descriptions, incorrect! No title in dbt models.
                "title": column.gooddata_ldm_title,
                "description": column.gooddata_ldm_description,
                "source_column": column.name,
                "source_column_data_type": column.data_type,
                "tags": [table.gooddata_ldm_title] + column.tags,
            }
            for column in table.columns.values()
            if column.gooddata_is_fact()
        ]
        return facts

    @staticmethod
    def make_labels(table: DbtModelTable, attribute_column: DbtModelColumn) -> tuple[list[dict], Optional[dict]]:
        labels = []
        default_view = None
        for column in table.columns.values():
            if column.gooddata_is_label(attribute_column.name):
                labels.append(
                    {
                        "id": column.ldm_id,
                        "title": column.gooddata_ldm_title,
                        "description": column.gooddata_ldm_description,
                        "source_column": column.name,
                        "source_column_data_type": column.data_type,
                        "value_type": column.label_type,
                        "tags": [table.gooddata_ldm_title] + column.tags,
                    }
                )
                if column.meta.gooddata.default_view:
                    default_view = {
                        "id": column.ldm_id,
                        "type": "label",
                    }
        return labels, default_view

    def make_attributes(self, table: DbtModelTable) -> list[dict]:
        attributes = []
        for column in table.columns.values():
            # Default is attribute
            if column.gooddata_is_attribute():
                labels, default_view = self.make_labels(table, column)
                attributes.append(
                    {
                        "id": column.ldm_id,
                        "title": column.gooddata_ldm_title,
                        "description": column.gooddata_ldm_description,
                        "source_column": column.name,
                        "source_column_data_type": column.data_type,
                        "tags": [table.gooddata_ldm_title] + column.tags,
                        "labels": labels,
                        "sort_column": column.meta.gooddata.sort_column,
                        "sort_direction": column.sort_direction,
                        "default_view": default_view,
                    }
                )
        return attributes

    @staticmethod
    def make_date_datasets(table: DbtModelTable, existing_date_datasets: list[dict]) -> list[dict]:
        date_datasets = []
        for column in table.columns.values():
            existing_dataset_ids = [d["id"] for d in existing_date_datasets]
            if column.is_date() and column.gooddata_ldm_id not in existing_dataset_ids:
                if column.data_type in TIMESTAMP_DATA_TYPES:
                    granularities = DATE_GRANULARITIES + TIMESTAMP_GRANULARITIES
                else:
                    granularities = DATE_GRANULARITIES
                date_datasets.append(
                    {
                        "id": column.ldm_id,
                        "title": column.gooddata_ldm_title,
                        "description": column.description,
                        # Put them all to a dedicated tag(folder).
                        # Date dims are not put to the corresponding folder in our Analytical Designer.
                        # Also, the same date dim can be created from multiple columns (shared date dims).
                        # So the tagging is useless here now, we should think about it in the future.
                        "tags": ["Date dimensions"],
                        "granularities": granularities,
                        "granularities_formatting": {
                            "title_base": "",
                            "title_pattern": "%titleBase - %granularityTitle",
                        },
                    }
                )
        return date_datasets

    def make_dataset(self, data_source_id: str, table: DbtModelTable, role_playing_tables: dict, result: dict) -> dict:
        grain = self.make_grain(table)
        references = self.make_references(table, role_playing_tables)
        facts = self.make_facts(table)
        attributes = self.make_attributes(table)

        result["datasets"].append(
            {
                "id": table.gooddata_ldm_id,
                "title": table.gooddata_ldm_title,
                "description": table.description,
                "tags": [table.gooddata_ldm_title] + table.tags,
                "data_source_table_id": {
                    "data_source_id": data_source_id,
                    # TODO - add support for multiple schemas
                    #  would have to be reflected in LDM IDs too
                    "id": table.name,
                    "path": [self.schema_name, table.name],
                    "type": "dataSource",
                },
                "grain": grain,
                "references": references,
                "facts": facts,
                "attributes": attributes,
            }
        )

        date_datasets = self.make_date_datasets(table, result["date_instances"])
        result["date_instances"] = result["date_instances"] + date_datasets
        return result

    @staticmethod
    def populate_role_playing_tables(tables: list[DbtModelTable], role_playing_tables: dict) -> list[DbtModelTable]:
        result = []
        for table in tables:
            if table.name in role_playing_tables:
                for role_column in role_playing_tables[table.name]:
                    new_table = copy.deepcopy(table)
                    new_table.role_name = role_column
                    for column in new_table.columns.values():
                        column.role_name = role_column
                    result.append(new_table)
            else:
                result.append(table)
        return result

    def make_declarative_datasets(self, data_source_id: str, model_ids: Optional[list[str]]) -> dict:
        result: dict[str, list] = {"datasets": [], "date_instances": []}
        model_tables = [t for t in self.tables if not model_ids or t.meta.gooddata.model_id in model_ids]
        role_playing_tables = self.find_role_playing_tables(model_tables)
        model_tables_with_roles = self.populate_role_playing_tables(model_tables, role_playing_tables)

        for table in model_tables_with_roles:
            result = self.make_dataset(data_source_id, table, role_playing_tables, result)
        return result

    def get_entity_type(self, table_name: str, column_name: str) -> Optional[str]:
        comp_table_name = table_name
        if self.upper_case:
            comp_table_name = table_name.upper()
        comp_column_name = column_name
        if self.upper_case:
            comp_column_name = column_name.upper()
        for table in self.tables:
            if table.name == comp_table_name:
                for column in table.columns.values():
                    if column.name == comp_column_name:
                        return column.ldm_type
        return None



# (C) 2023 GoodData Corporation
import logging


def get_logger(name: str, debug: bool = False) -> logging.Logger:
    level = logging.INFO if not debug else logging.DEBUG
    logging.basicConfig(level=level, format="%(levelname)-8s: %(name)s : %(asctime)-15s - %(message)s")
    logger = logging.getLogger(name)
    return logger
